# File for defining how to form a scheudle consisting of heuristics and time bounds.
# When a schedule is implemented it must be added to the "schedules" dict such that
# there can be a mapping between its string and the specification
#
# The general format of a schedule is: sched = [[h0, 120], [h2, 300]]
# where h0, h2 are heuristics and 120, 300 is the time to run it for.
# If None is supplied as time, the heuristic runs indefinitely.


default = [("heur/default", None)]
default_test = [("heur/default", 2), ("heur/default", None)]
default_claus_test = [("heur/default_claus_test", 2), ("heur/default_claus_test", None)]

test_parallel = [[('heur/471148', 20)],
                 [('heur/469570', 4), ('heur/330959', 16)]]


ueq_schedule = [[('heur/471148', 300)],
                [('heur/409472', 7), ('heur/397416', 30), ('heur/492340', 119), ('heur/346298', 144)],
                [('heur/346304', 17), ('heur/346301', 283)],
                [('heur/397445', 1), ('heur/397453', 106), ('heur/379313', 193)],
                [('heur/330959', 5), ('heur/379314', 97), ('heur/409461', 198)],
                [('heur/397448', 8), ('heur/421501', 9), ('heur/397449', 29), ('heur/346297', 58), ('heur/409477', 196)],
                [('heur/492341', 300)],
                [('heur/409474', 1), ('heur/397450', 8), ('heur/409466', 18), ('heur/457559', 24), ('heur/421509', 249)]]

ltb_fof_schedule = [[('heur/471148', 300)],
                    [('heur/471148_sine_1_1', 300)],
                    [('heur/381325', 1), ('heur/381332', 14), ('heur/346291', 285)],
                    [('heur/346295', 2), ('heur/379308', 14), ('heur/409476', 18), ('heur/492340', 83), ('heur/421509', 183)],
                    [('heur/409474', 1), ('heur/397438', 3), ('heur/346294', 6), ('heur/397433', 6), ('heur/409466', 6), ('heur/379305', 8), ('heur/397452', 11), ('heur/492338', 13), ('heur/433536', 17), ('heur/397454', 18), ('heur/492341', 211)],
                    [('heur/346297', 1), ('heur/421498', 1), ('heur/469572', 1), ('heur/379306', 2), ('heur/379320', 2), ('heur/409461', 2), ('heur/421501', 2), ('heur/346293', 3), ('heur/445550', 16), ('heur/421494', 20), ('heur/379314', 35), ('heur/492343', 72), ('heur/469574', 143)],
                    [('heur/397453', 2), ('heur/445545', 2), ('heur/330959', 296)],
                    [('heur/397423', 2), ('heur/346296', 4), ('heur/346305', 4), ('heur/397442', 4), ('heur/469570', 4), ('heur/445546', 5), ('heur/421528', 8), ('heur/381331', 10), ('heur/397437', 16), ('heur/457559', 24), ('heur/346301', 29), ('heur/379313', 46), ('heur/346292', 59), ('heur/469569', 85)]]



ltb_super_22_6C = [[('heur/471148', 300)],
                   [('heur/471148_sine_1_1', 300)],
                   [('heur/381325', 1), ('heur/469574', 9), ('heur/346292', 12)],
                   [('heur/381326', 1), ('heur/381332', 2), ('heur/469568', 2), ('heur/492341', 17)],
                   [('heur/346299', 1), ('heur/379313', 2), ('heur/409491', 19)],
                   [('heur/346301', 1), ('heur/379311', 1), ('heur/397444', 1), ('heur/433534', 19)],
                   [('heur/346293', 1), ('heur/492338', 21)],
                   [('heur/381328', 2), ('heur/381333', 20)]]

ltb_super_22_5C = [[('heur/471148', 300)],
                   [('heur/471148_sine_1_1', 300)],
                   [('heur/471148_sine_1.2_3', 300)],
                       [('heur/346301', 1), ('heur/346292', 21)],
                   [('heur/346293', 1), ('heur/381325', 1), ('heur/492338', 20)],
                   [('heur/492341', 22)],
                   [('heur/379312', 1), ('heur/381326', 1), ('heur/409466', 1), ('heur/381328', 2), ('heur/433535', 2), ('heur/409476', 15)],
                   [('heur/379313', 1), ('heur/469569', 21)]]


ltb_low_phase_v4 = [('heur/280924', 12),
                    ('heur/277968', 12),
                    ('heur/277969', 12),
                    ('heur/278164', 11),
                    ('heur/278644', 11),
                    ('heur/280746', 10),
                    ('heur/260291', 12),
                    ('heur/278359', 7),
                    ('heur/258434', 4),
                    ('heur/279708', 1),
                    ('heur/264563', 2),
                    ('heur/280340', 1),
                    ('heur/259599', 1)]



ltb_low_phase_v5 = [('heur/269506', 12),
                    ('heur/281151', 5),
                    ('heur/266513', 11),
                    ('heur/268297', 12),
                    ('heur/279169', 3),
                    ('heur/278555', 7),
                    ('heur/278359', 8),
                    ('heur/278644', 10),
                    ('heur/278166', 8),
                    ('heur/281331', 9),
                    ('heur/277960', 5),
                    ('heur/279708', 2),
                    ('heur/269504', 2),
                    ('heur/269566', 1),
                    ('heur/265544', 1)]


#ltb_high_common = [('heur/277962', 14), ('heur/277959', 163)]
ltb_high_common = [('heur/280746', 14), ('heur/277959', 163)]




ltb_high_phase_schedule = [('heur/261846', 13), ('heur/259793', 167)]


fnt_schedule = [('heur/fnt_1', 300),
                ('heur/fnt_2', 300),
                ('heur/fnt_3', 300),
                ('heur/fnt_4', 300),
                ('heur/271307', 300),
                ('heur/273393', 11),
                ('heur/271303', 26),
                ('heur/269879', 39),
                ('heur/271021', 43),
                ('heur/272218', 59),
                ('heur/273030', 90),
                ('heur/270906', 110),
                ('heur/272766', 137)]


fof_schedule = [[('heur/471148', 300)],
                [('heur/409474', 1), ('heur/397453', 2), ('heur/421501', 2), ('heur/445550', 16), ('heur/409476', 18), ('heur/379313', 79), ('heur/421509', 182)],
                [('heur/469571', 1), ('heur/397438', 3), ('heur/330959', 296)],
                [('heur/469570', 4), ('heur/397433', 6), ('heur/421523', 13), ('heur/381332', 14), ('heur/397437', 16), ('heur/421494', 20), ('heur/469569', 25), ('heur/346292', 59), ('heur/469574', 143)],
                [('heur/346296', 4), ('heur/457559', 296)],
                [('heur/381325', 1), ('heur/397442', 1), ('heur/346295', 2), ('heur/379306', 2), ('heur/379320', 2), ('heur/389350', 5), ('heur/346294', 6), ('heur/492343', 6), ('heur/397452', 11), ('heur/379308', 12), ('heur/469576', 14), ('heur/381331', 15), ('heur/397454', 17), ('heur/433536', 17), ('heur/409466', 18), ('heur/492338', 24), ('heur/346301', 29), ('heur/379314', 35), ('heur/492340', 83)],
                [('heur/346297', 1), ('heur/421498', 1), ('heur/445545', 2), ('heur/346293', 3), ('heur/421528', 8), ('heur/492341', 285)],
                [('heur/397423', 2), ('heur/346305', 4), ('heur/445546', 5), ('heur/346291', 289)]]


sup_cicm_20_train_default = [('heur/381325', 1), ('heur/389408', 19)]


sup_cicm_20_train_parallel = [[('heur/381325', 20)],
                              [('heur/389408', 19)]]

sup_cicm_20_train_admissible = {"default": sup_cicm_20_train_default,
                                "model": "embeddings/embedding_2021-04-26_11-32-34.log.pkl",
                                "cluster_centers": [[1,1,0,0,0,1,1,0,0,0,0,1,1,1,0,0,1,1,1,1,1,1,1,0,1,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,0,0,1,1,1,1,1,1,1],
                                                    [0,0,1,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,1]],
                                0: [('heur/381325', 1), ('heur/383337', 2), ('heur/381333', 17)],
                                1: [('heur/379319', 3), ('heur/389408', 17)]
                                }

test_single_core_300_super = [[('heur/346294', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/379312', 1), ('heur/381325', 1), ('heur/409491', 1), ('heur/433536', 1),
                              ('heur/346293', 2), ('heur/346295', 2), ('heur/379308', 2), ('heur/379311', 2), ('heur/381332', 2), ('heur/397423', 2), ('heur/379320', 3),
                              ('heur/397453', 3), ('heur/379313', 5), ('heur/379314', 5), ('heur/381331', 6), ('heur/469568', 6), ('heur/469574', 7), ('heur/433535', 12),
                              ('heur/457559', 12), ('heur/346292', 65), ('heur/471148', 157)]]

test_single_core_300_super_x2 = [[('heur/346294', 2), ('heur/346299', 2), ('heur/346301', 2), ('heur/379312', 2), ('heur/381325', 2), ('heur/409491', 2), ('heur/433536', 2), ('heur/346293', 4), ('heur/346295', 4), ('heur/379308', 4), ('heur/379311', 4), ('heur/381332', 4), ('heur/397423', 4), ('heur/379320', 6), ('heur/397453', 6), ('heur/379313', 10), ('heur/379314', 10), ('heur/381331', 12), ('heur/469568', 12), ('heur/469574', 14), ('heur/433535', 24), ('heur/457559', 24), ('heur/346292', 130), ('heur/471148', 314)]]


test_single_core_300_super_x3 = [[('heur/346294', 3), ('heur/346299', 3), ('heur/346301', 3), ('heur/379312', 3), ('heur/381325', 3), ('heur/409491', 3), ('heur/433536', 3), ('heur/346293', 6), ('heur/346295', 6), ('heur/379308', 6), ('heur/379311', 6), ('heur/381332', 6), ('heur/397423', 6), ('heur/379320', 9), ('heur/397453', 9), ('heur/379313', 15), ('heur/379314', 15), ('heur/381331', 18), ('heur/469568', 18), ('heur/469574', 21), ('heur/433535', 36), ('heur/457559', 36), ('heur/346292', 195), ('heur/471148', 471)]]


test_single_core_300_super_x4 = [[('heur/346294', 4), ('heur/346299', 4), ('heur/346301', 4), ('heur/379312', 4), ('heur/381325', 4), ('heur/409491', 4), ('heur/433536', 4), ('heur/346293', 8), ('heur/346295', 8), ('heur/379308', 8), ('heur/379311', 8), ('heur/381332', 8), ('heur/397423', 8), ('heur/379320', 12), ('heur/397453', 12), ('heur/379313', 20), ('heur/379314', 20), ('heur/381331', 24), ('heur/469568', 24), ('heur/469574', 28), ('heur/433535', 48), ('heur/457559', 48), ('heur/346292', 260), ('heur/471148', 628)]]

test_single_core_300_mixed = [('heur/346294', 1), ('heur/346299', 1), ('heur/379307', 1), ('heur/379312', 1), ('heur/381325', 1), ('heur/381326', 1),
                              ('heur/397442', 1), ('heur/409466', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/421498', 1), ('heur/346293', 2),
                              ('heur/346295', 2), ('heur/379311', 2), ('heur/381332', 2), ('heur/386344', 2), ('heur/379320', 3), ('heur/379308', 4),
                              ('heur/421528', 4), ('heur/379313', 5), ('heur/379314', 5), ('heur/397453', 5), ('heur/469574', 7), ('heur/457559', 12),
                              ('heur/433534', 20), ('heur/469569', 25), ('heur/346292', 65), ('heur/389408', 124)]

test_single_core_300 = [('heur/379312', 1), ('heur/381322', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397452', 1), ('heur/409466', 1),
                        ('heur/409482', 1), ('heur/433535', 1), ('heur/469573', 1), ('heur/397423', 2), ('heur/421511', 2), ('heur/469576', 2),
                        ('heur/397453', 3), ('heur/421528', 4), ('heur/469570', 4), ('heur/381331', 5), ('heur/379314', 11), ('heur/409477', 14),
                        ('heur/421523', 14), ('heur/397416', 17), ('heur/397437', 17), ('heur/421531', 18), ('heur/379313', 19), ('heur/457559', 19),
                        ('heur/389408', 20), ('heur/397454', 20), ('heur/409476', 20), ('heur/421513', 20), ('heur/433534', 20), ('heur/469569', 20),
                        ('heur/469574', 20)]

test_single_core_300_admissible = {"default": test_single_core_300_mixed,
                                   "model": "embeddings/embedding_2021-05-06_18-13-03.log.pkl",
                                   "cluster_centers": [[0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                                       [1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                                        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
                                   0: [('heur/421514', 1), ('heur/433535', 1), ('heur/421510', 2), ('heur/433536', 3), ('heur/433541', 6), ('heur/397452', 7), ('heur/421515', 10), ('heur/379314', 13), ('heur/397451', 15), ('heur/397423', 16), ('heur/409477', 16), ('heur/397436', 17), ('heur/379313', 18), ('heur/409466', 18), ('heur/379312', 19), ('heur/421526', 19), ('heur/433537', 19), ('heur/389408', 20), ('heur/397453', 20), ('heur/421509', 20), ('heur/421513', 20), ('heur/433534', 20)],
                                   1: [('heur/379314', 1), ('heur/381322', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/409465', 1), ('heur/409474', 1), ('heur/409482', 1), ('heur/469573', 1), ('heur/421501', 2), ('heur/421511', 2), ('heur/397438', 3), ('heur/397442', 4), ('heur/421528', 4), ('heur/469570', 4), ('heur/469578', 5), ('heur/421505', 8), ('heur/381331', 10), ('heur/397429', 10), ('heur/397449', 14), ('heur/421523', 14), ('heur/397416', 17), ('heur/397437', 17), ('heur/409489', 19), ('heur/457559', 19), ('heur/397454', 20), ('heur/409476', 20), ('heur/409479', 20), ('heur/445542', 20), ('heur/445550', 20), ('heur/469569', 20), ('heur/469574', 20)]}


test_single_core_300_admissible_super = {"default": test_single_core_300_super,
                                         #"model": "embeddings/embedding_2021-05-16_20-52-31.log.pkl",
                                         #"model": "embeddings/embedding_2021-05-21_16-52-56.log.pkl",
                                         "model": "embeddings/embedding_2021-05-23_15-55-26.log.pkl",
                                   "cluster_centers":[[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
[0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],

0 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
1 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
2 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
3 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
4 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
5 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
6 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
7 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
8 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
9 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
10 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
11 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
12 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
13 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
14 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
15 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
16 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
17 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
18 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
19 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
20 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
21 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
22 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
23 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
24 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
25 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
26 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
27 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
28 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
29 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
30 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
31 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
32 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
33 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
34 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
35 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
36 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
37 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
38 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
39 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
40 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
41 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
42 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
43 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
44 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
45 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
46 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
47 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
48 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
49 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
50 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
51 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
52 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
53 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
54 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
55 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
56 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
57 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
58 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
59 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
60 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
61 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
62 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
63 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
64 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
65 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
66 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
67 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
68 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
69 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
70 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
71 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
72 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
73 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
74 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
75 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
76 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
77 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
78 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
79 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
80 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
81 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
82 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
83 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
84 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
85 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
86 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
87 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
88 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
89 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
90 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
91 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
92 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
93 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
94 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
95 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
96 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
97 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
98 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)],
99 : [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/433535', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379304', 9), ('heur/330959', 25), ('heur/379314', 48), ('heur/346296', 155)]}



smt_4000_super_300_4 = [[('heur/397433', 2), ('heur/421501', 2), ('heur/445545', 2), ('heur/381332', 4), ('heur/397418', 4), ('heur/397452', 7), ('heur/445550', 8), ('heur/346301', 12), ('heur/421509', 46), ('heur/469569', 85), ('heur/389408', 128)],
                        [('heur/381325', 1), ('heur/409474', 1), ('heur/433536', 1), ('heur/469572', 1), ('heur/346293', 2), ('heur/346296', 3), ('heur/386344', 3), ('heur/346294', 6), ('heur/469576', 6), ('heur/469574', 23), ('heur/457559', 25), ('heur/346292', 228)],
                        [('heur/346297', 1), ('heur/421498', 1), ('heur/433535', 1), ('heur/397423', 2), ('heur/379320', 3), ('heur/397438', 3), ('heur/397453', 4), ('heur/379311', 5), ('heur/409466', 6), ('heur/379308', 15), ('heur/379305', 16), ('heur/409476', 18), ('heur/433537', 20), ('heur/379313', 46), ('heur/346291', 61), ('heur/379314', 98)],
                        [('heur/346295', 2), ('heur/381331', 10), ('heur/471148', 288)]]



smt_4000_super_single_600 = [('heur/346294', 1), ('heur/346297', 1), ('heur/346301', 1), ('heur/381325', 1), ('heur/381330', 1), ('heur/383336', 1), ('heur/397416', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/409474', 1), ('heur/421498', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/469573', 1), ('heur/346293', 2), ('heur/346295', 2), ('heur/379311', 2), ('heur/381332', 2), ('heur/397423', 2), ('heur/397450', 2), ('heur/421501', 2), ('heur/469576', 2), ('heur/379320', 3), ('heur/379313', 5), ('heur/379314', 5), ('heur/397453', 5), ('heur/469568', 5), ('heur/445550', 8), ('heur/381331', 10), ('heur/457559', 12), ('heur/379308', 15), ('heur/409477', 17), ('heur/433537', 20), ('heur/469574', 23), ('heur/346292', 161), ('heur/471148', 281)]


smt_4000_super_single_900 = [('heur/330959', 1), ('heur/346294', 1), ('heur/346297', 1), ('heur/381325', 1), ('heur/397444', 1), ('heur/409474', 1), ('heur/421498', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/469573', 1), ('heur/346293', 2), ('heur/379311', 2), ('heur/397423', 2), ('heur/397433', 2), ('heur/397442', 2), ('heur/409461', 2), ('heur/421501', 2), ('heur/445545', 2), ('heur/379320', 3), ('heur/397438', 3), ('heur/469576', 3), ('heur/346295', 4), ('heur/397418', 4), ('heur/397452', 4), ('heur/397453', 4), ('heur/346291', 5), ('heur/409466', 6), ('heur/445550', 8), ('heur/379314', 9), ('heur/346301', 12), ('heur/381331', 12), ('heur/381332', 14), ('heur/379308', 15), ('heur/409477', 16), ('heur/397454', 17), ('heur/409476', 18), ('heur/433537', 20), ('heur/469574', 23), ('heur/457559', 25), ('heur/421509', 40), ('heur/379313', 46), ('heur/469569', 76), ('heur/346292', 200), ('heur/471148', 287)]


smt_4000_super_single_1200 = [('heur/346297', 1), ('heur/381325', 1), ('heur/409474', 1), ('heur/421498', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/469572', 1), ('heur/346293', 2), ('heur/346295', 2), ('heur/379306', 2), ('heur/397423', 2), ('heur/397433', 2), ('heur/421501', 2), ('heur/445545', 2), ('heur/346296', 3), ('heur/379320', 3), ('heur/386344', 3), ('heur/397438', 3), ('heur/397418', 4), ('heur/397453', 4), ('heur/346294', 6), ('heur/409466', 6), ('heur/469576', 6), ('heur/397452', 7), ('heur/445550', 8), ('heur/381331', 10), ('heur/346301', 12), ('heur/381332', 14), ('heur/379308', 15), ('heur/409477', 16), ('heur/409476', 18), ('heur/433537', 20), ('heur/469574', 23), ('heur/457559', 25), ('heur/379313', 46), ('heur/421509', 46), ('heur/346291', 61), ('heur/469569', 85), ('heur/379314', 98), ('heur/389408', 128), ('heur/346292', 228), ('heur/471148', 281)]


smt_3scp_default_1200 = [[('heur/default', 1200)],
                         [('heur/346297', 1), ('heur/381325', 1), ('heur/379320', 3), ('heur/346305', 4), ('heur/397450', 7), ('heur/397418', 10), ('heur/386344', 11), ('heur/346294', 13), ('heur/381332', 14), ('heur/409466', 18), ('heur/469569', 93), ('heur/469574', 159), ('heur/346304', 169), ('heur/379314', 172), ('heur/379313', 228), ('heur/457559', 297)],
                         [('heur/409474', 1), ('heur/469573', 1), ('heur/346296', 3), ('heur/469575', 5), ('heur/421528', 8), ('heur/421523', 14), ('heur/445550', 16), ('heur/381331', 39), ('heur/330959', 236), ('heur/346291', 289), ('heur/346292', 292), ('heur/471148', 296)],
                         [('heur/433535', 1), ('heur/346295', 2), ('heur/445545', 2), ('heur/383335', 4), ('heur/379311', 5), ('heur/397433', 7), ('heur/397452', 7), ('heur/421506', 8), ('heur/379312', 9), ('heur/421512', 9), ('heur/346303', 10), ('heur/379308', 14), ('heur/397454', 17), ('heur/421494', 17), ('heur/433536', 17), ('heur/386347', 18), ('heur/409476', 18), ('heur/433537', 18), ('heur/346301', 98), ('heur/397453', 162), ('heur/421509', 199), ('heur/397431', 258), ('heur/409477', 300)]]


smt_3scp_default_300_x4 = [[('heur/default', 1200)],
                           [('heur/346294', 24), ('heur/346301', 104), ('heur/421509', 160), ('heur/346292', 912)],
                           [('heur/330959', 4), ('heur/346297', 4), ('heur/381325', 4), ('heur/397444', 4), ('heur/409474', 4), ('heur/433535', 4), ('heur/433536', 4), ('heur/469571', 4), ('heur/346293', 8), ('heur/397423', 8), ('heur/397433', 8), ('heur/397442', 8), ('heur/397450', 8), ('heur/409461', 8), ('heur/421501', 8), ('heur/445545', 8), ('heur/346296', 12), ('heur/379320', 12), ('heur/386344', 12), ('heur/469576', 12), ('heur/346295', 16), ('heur/381332', 16), ('heur/397418', 16), ('heur/397452', 16), ('heur/346291', 20), ('heur/379319', 20), ('heur/397453', 20), ('heur/409466', 24), ('heur/445550', 32), ('heur/379313', 48), ('heur/381331', 48), ('heur/379308', 60), ('heur/379305', 64), ('heur/397454', 68), ('heur/433537', 80), ('heur/469574', 92), ('heur/457559', 100), ('heur/469569', 316)],
                           [('heur/421498', 4), ('heur/397438', 12), ('heur/379314', 36), ('heur/471148', 1148)]]



smt_2scp_default_with_sat_1200 = [[('heur/default', 1200)],
                                  [('heur/smt_with_sat', 1200)],
                                  [('heur/381325', 1), ('heur/397423', 1), ('heur/469572', 1), ('heur/346295', 2), ('heur/445545', 2), ('heur/379320', 3), ('heur/346296', 4), ('heur/346305', 4), ('heur/346293', 5), ('heur/379306', 5), ('heur/445546', 5), ('heur/397437', 6), ('heur/397433', 7), ('heur/421506', 8), ('heur/421528', 8), ('heur/421501', 9), ('heur/346294', 13), ('heur/379308', 14), ('heur/381332', 14), ('heur/421523', 14), ('heur/381331', 15), ('heur/409471', 16), ('heur/433536', 17), ('heur/386347', 18), ('heur/409466', 18), ('heur/433537', 18), ('heur/409479', 20), ('heur/421494', 20), ('heur/346292', 51), ('heur/457559', 97), ('heur/421509', 199), ('heur/346291', 289), ('heur/330959', 296)],
                                  [('heur/346297', 1), ('heur/409474', 1), ('heur/421498', 1), ('heur/433535', 1), ('heur/346298', 4), ('heur/383335', 4), ('heur/469570', 4), ('heur/346302', 7), ('heur/397450', 7), ('heur/397452', 7), ('heur/397418', 10), ('heur/386344', 11), ('heur/445550', 16), ('heur/469575', 16), ('heur/397454', 17), ('heur/409476', 18), ('heur/409477', 41), ('heur/397453', 70), ('heur/469569', 85), ('heur/346301', 98), ('heur/379314', 98), ('heur/469574', 159),
                                 ('heur/379313', 228), ('heur/471148', 296)]]

smt_2scp_default_with_sat_300_x4 = [[('heur/default', 1200)],
                                    [('heur/smt_with_sat', 1200)],
                                    [('heur/389351', 4), ('heur/421498', 4), ('heur/346293', 8), ('heur/379320', 12), ('heur/457559', 48), ('heur/471148', 1124)],
                                    [('heur/346294', 4), ('heur/346297', 4), ('heur/346301', 4), ('heur/379312', 4), ('heur/381325', 4), ('heur/383340', 4), ('heur/397441', 4), ('heur/397444', 4), ('heur/397452', 4), ('heur/409474', 4), ('heur/433535', 4), ('heur/433536', 4), ('heur/469573', 4), ('heur/346295', 8), ('heur/379311', 8), ('heur/381332', 8), ('heur/397423', 8), ('heur/397450', 8), ('heur/421501', 8), ('heur/469576', 8), ('heur/379313', 20), ('heur/379314', 20), ('heur/397453', 20), ('heur/469568', 20), ('heur/445550', 32), ('heur/381331', 48), ('heur/379305', 56), ('heur/379308', 60), ('heur/433537', 80), ('heur/469574', 92), ('heur/346292', 644)]]




admissible_schedule_test_recomputed = {"default": test_single_core_300_super,
                                       "model": "embeddings/embedding_2021-05-23_15-55-26.log.pkl",
                                       'cluster_centers': [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0],
[0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
[0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
0: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346298', 7), ('heur/397422', 8), ('heur/379309', 101)],
1: [('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 11), ('heur/346291', 33), ('heur/346293', 54)],
2: [('heur/346292', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397451', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409490', 1), ('heur/421507', 1), ('heur/421514', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/346291', 11), ('heur/346293', 17), ('heur/346294', 208)],
3: [('heur/330959', 1), ('heur/346291', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433531', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 7), ('heur/379307', 7), ('heur/346292', 21), ('heur/346295', 103)],
4: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409489', 1), ('heur/409491', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379308', 7), ('heur/421509', 10), ('heur/409477', 50), ('heur/346294', 97)],
5: [('heur/471148', 300)],
6: [('heur/346296', 11), ('heur/397416', 69), ('heur/346291', 102), ('heur/379314', 118)],
7: [('heur/346292', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/381321', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346297', 5), ('heur/381334', 6), ('heur/346295', 9), ('heur/346294', 10), ('heur/346293', 14), ('heur/346302', 138)],
8: [('heur/346294', 300)],
9: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397447', 1), ('heur/397451', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421526', 1), ('heur/421530', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469569', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/346293', 210)],
10: [('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346305', 5), ('heur/346298', 17), ('heur/330959', 18), ('heur/346294', 54), ('heur/346297', 117)],
11: [('heur/330959', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346298', 5), ('heur/346292', 43), ('heur/346291', 186)],
12: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 136)],
13: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381321', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433536', 1), ('heur/445547', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346298', 175)],
14: [('heur/346292', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 5), ('heur/346300', 5), ('heur/346295', 7), ('heur/346296', 11), ('heur/346293', 20), ('heur/346294', 96)],
15: [('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/379306', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379319', 1), ('heur/381325', 1), ('heur/383340', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397451', 1), ('heur/409462', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409478', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/381332', 263)],
16: [('heur/346298', 1), ('heur/346301', 1), ('heur/381332', 1), ('heur/397444', 1), ('heur/409463', 1), ('heur/409477', 1), ('heur/421513', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/346305', 5), ('heur/379308', 5), ('heur/346295', 274)],
17: [('heur/330959', 1), ('heur/346291', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346303', 5), ('heur/346292', 15), ('heur/346293', 95)],
18: [('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 21), ('heur/330959', 24), ('heur/346297', 31), ('heur/346296', 38)],
19: [('heur/379313', 300)],
20: [('heur/409467', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346298', 7), ('heur/330959', 18), ('heur/346296', 23), ('heur/346293', 28), ('heur/346291', 151)],
21: [('heur/469572', 1), ('heur/469573', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346292', 7), ('heur/379308', 8), ('heur/346295', 66), ('heur/346296', 67), ('heur/330959', 148)],
22: [('heur/381331', 1), ('heur/381332', 1), ('heur/383340', 1), ('heur/386346', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397450', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421509', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 21), ('heur/346291', 228)],
23: [('heur/330959', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/457559', 283)],
24: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397418', 1), ('heur/397424', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397423', 202)],
25: [('heur/386346', 1), ('heur/409467', 1), ('heur/409476', 1), ('heur/409482', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421498', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/379306', 5), ('heur/346294', 7), ('heur/346297', 8), ('heur/346305', 8), ('heur/346292', 9), ('heur/346295', 10), ('heur/445546', 11), ('heur/346296', 21), ('heur/346298', 24), ('heur/330959', 33), ('heur/346300', 45), ('heur/397416', 105)],
26: [('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 8), ('heur/346300', 9), ('heur/346292', 15), ('heur/346291', 39), ('heur/346298', 176)],
27: [('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/381333', 7), ('heur/346291', 100)],
28: [('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 5), ('heur/346300', 5), ('heur/346293', 7), ('heur/381325', 11), ('heur/330959', 16), ('heur/346297', 18), ('heur/346295', 23), ('heur/346292', 25), ('heur/346291', 183)],
29: [('heur/330959', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 6), ('heur/346295', 6), ('heur/346291', 130)],
30: [('heur/346300', 300)],
31: [('heur/330959', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 30), ('heur/346295', 254)],
32: [('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346291', 14), ('heur/330959', 175)],
33: [('heur/346302', 300)],
34: [('heur/346296', 1), ('heur/346305', 1), ('heur/379313', 1), ('heur/386346', 1), ('heur/397444', 1), ('heur/421531', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/346294', 292)],
35: [('heur/330959', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346294', 8), ('heur/346293', 32), ('heur/346291', 56), ('heur/409461', 174)],
36: [('heur/346291', 1), ('heur/346292', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/381324', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/330959', 211)],
37: [('heur/379314', 300)],
38: [('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346298', 8), ('heur/346300', 8), ('heur/346303', 14), ('heur/346291', 61), ('heur/330959', 106)],
39: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/397416', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346299', 5), ('heur/346298', 36), ('heur/346291', 134)],
40: [('heur/421525', 8), ('heur/346300', 35), ('heur/346292', 257)],
41: [('heur/421531', 1), ('heur/433529', 1), ('heur/346304', 6), ('heur/346291', 7), ('heur/346297', 10), ('heur/346301', 17), ('heur/346299', 84), ('heur/330959', 174)],
42: [('heur/346305', 5), ('heur/381325', 7), ('heur/379304', 9), ('heur/346293', 10), ('heur/379306', 11), ('heur/379313', 11), ('heur/379309', 12), ('heur/330959', 25), ('heur/346301', 41), ('heur/346299', 51), ('heur/346302', 54), ('heur/346303', 64)],
43: [('heur/346294', 5), ('heur/346299', 5), ('heur/381325', 6), ('heur/379306', 9), ('heur/379316', 11), ('heur/397452', 11), ('heur/346292', 16), ('heur/346304', 22), ('heur/346296', 25), ('heur/330959', 79), ('heur/379314', 111)],
44: [('heur/330959', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 274)],
45: [('heur/346298', 300)],
46: [('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/421509', 286)],
47: [('heur/397451', 12), ('heur/379313', 54), ('heur/346297', 234)],
48: [('heur/346292', 1), ('heur/421496', 1), ('heur/421514', 1), ('heur/421521', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445545', 1), ('heur/457557', 1), ('heur/457565', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/346304', 286)],
49: [('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397423', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397451', 1), ('heur/409457', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445545', 1), ('heur/457559', 1), ('heur/469579', 1), ('heur/379306', 9), ('heur/346297', 15), ('heur/346295', 46), ('heur/346293', 179)],
50: [('heur/330959', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/469574', 242)],
51: [('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 7), ('heur/346303', 8), ('heur/346300', 13), ('heur/346291', 226)],
52: [('heur/471148', 1), ('heur/409477', 299)],
53: [('heur/397452', 300)],
54: [('heur/409478', 7), ('heur/381332', 293)],
55: [('heur/381322', 8), ('heur/346291', 79), ('heur/346294', 213)],
56: [('heur/346294', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381325', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/386347', 1), ('heur/386350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397417', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/409460', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445554', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469574', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346295', 5), ('heur/346296', 7), ('heur/346292', 15), ('heur/346293', 190)],
57: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/379308', 1), ('heur/381332', 1), ('heur/397451', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/445546', 1), ('heur/346305', 286)],
58: [('heur/346292', 1), ('heur/346299', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379313', 1), ('heur/379315', 1), ('heur/379320', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/389350', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409482', 1), ('heur/409486', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/346297', 7), ('heur/346291', 12), ('heur/346293', 163)],
59: [('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379311', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397428', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/409462', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/457558', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379306', 9), ('heur/346296', 31), ('heur/346292', 184)],
60: [('heur/330959', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/409466', 6), ('heur/346305', 286)],
61: [('heur/379310', 300)],
62: [('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346297', 293)],
63: [('heur/330959', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/421528', 291)],
64: [('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 7), ('heur/433540', 9), ('heur/330959', 21), ('heur/346291', 182)],
65: [('heur/346292', 300)],
66: [('heur/397450', 300)],
67: [('heur/330959', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/409461', 279)],
68: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381331', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/381332', 114)],
69: [('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/409468', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 11), ('heur/381328', 19), ('heur/346296', 21), ('heur/346298', 41), ('heur/330959', 107)],
70: [('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379305', 8), ('heur/346297', 16), ('heur/346292', 31), ('heur/330959', 60), ('heur/346291', 71), ('heur/346293', 91)],
71: [('heur/381325', 300)],
72: [('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346302', 139)],
73: [('heur/330959', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/381331', 286)],
74: [('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 11), ('heur/346292', 17), ('heur/346293', 142)],
75: [('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 11), ('heur/346296', 97), ('heur/330959', 108)],
76: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/389408', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409477', 1), ('heur/421509', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/471148', 1), ('heur/381331', 277)],
77: [('heur/346295', 1), ('heur/379309', 1), ('heur/379313', 1), ('heur/381323', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445554', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/346292', 7), ('heur/330959', 8), ('heur/379318', 8), ('heur/379308', 9), ('heur/386345', 9), ('heur/397416', 9), ('heur/397417', 9), ('heur/379305', 12), ('heur/383338', 12), ('heur/389408', 15), ('heur/346297', 18), ('heur/383335', 19), ('heur/421509', 40), ('heur/346299', 111)],
78: [('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409476', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346300', 6), ('heur/330959', 77), ('heur/346301', 113)],
79: [('heur/421494', 11), ('heur/421509', 289)],
80: [('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383338', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/386351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421510', 1), ('heur/421522', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457560', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/471148', 1), ('heur/346294', 6), ('heur/346300', 9), ('heur/330959', 28), ('heur/346291', 151)],
81: [('heur/421500', 1), ('heur/421501', 1), ('heur/379304', 10), ('heur/330959', 11), ('heur/346299', 31), ('heur/346298', 53), ('heur/346304', 75), ('heur/346297', 118)],
82: [('heur/381327', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/389350', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397436', 1), ('heur/397439', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/346291', 5), ('heur/346297', 5), ('heur/346295', 13), ('heur/346294', 187)],
83: [('heur/330959', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/409477', 228)],
84: [('heur/346294', 1), ('heur/346297', 1), ('heur/379313', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386347', 1), ('heur/397419', 1), ('heur/397424', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/421492', 1), ('heur/421502', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/457565', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/379308', 255)],
85: [('heur/469575', 16), ('heur/389408', 284)],
86: [('heur/330959', 1), ('heur/409467', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346301', 6), ('heur/346298', 11), ('heur/346299', 11), ('heur/379308', 11), ('heur/346294', 12), ('heur/379305', 12), ('heur/379313', 231)],
87: [('heur/409490', 1), ('heur/409491', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/379317', 5), ('heur/379309', 6), ('heur/379318', 8), ('heur/383335', 10), ('heur/383337', 12), ('heur/381327', 13), ('heur/386347', 14), ('heur/346301', 21), ('heur/379313', 25), ('heur/346300', 86), ('heur/330959', 96)],
88: [('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346292', 9), ('heur/346293', 51), ('heur/346299', 231)],
89: [('heur/421531', 9), ('heur/346302', 291)],
90: [('heur/397430', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397446', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409458', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/346298', 5), ('heur/346301', 11), ('heur/346303', 14), ('heur/346295', 15), ('heur/346296', 196)],
91: [('heur/330959', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 297)],
92: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383338', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/421513', 16), ('heur/330959', 124)],
93: [('heur/397454', 17), ('heur/381331', 102), ('heur/379314', 181)],
94: [('heur/330959', 1), ('heur/346297', 1), ('heur/346305', 1), ('heur/386350', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409488', 6), ('heur/397422', 15), ('heur/346303', 48), ('heur/346291', 50), ('heur/346292', 174)],
95: [('heur/330959', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397444', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409476', 1), ('heur/409482', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421496', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421514', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346296', 223)],
96: [('heur/386344', 300)],
97: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 104)],
98: [('heur/397452', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/421493', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445546', 1), ('heur/457557', 1), ('heur/457561', 1), ('heur/457567', 1), ('heur/469568', 1), ('heur/469579', 1), ('heur/379311', 5), ('heur/381325', 5), ('heur/346294', 115), ('heur/346296', 138)],
99: [('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/421494', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421514', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433540', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457565', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/381326', 5), ('heur/346295', 11), ('heur/346293', 14), ('heur/346292', 41), ('heur/346291', 184)],
100: [('heur/389408', 300)],
101: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397430', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409466', 1), ('heur/409469', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445545', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/346305', 231)],
102: [('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 5), ('heur/346291', 12), ('heur/330959', 44)],
103: [('heur/330959', 1), ('heur/346301', 299)],
104: [('heur/421494', 1), ('heur/379308', 6), ('heur/379317', 7), ('heur/346298', 8), ('heur/379304', 8), ('heur/379320', 8), ('heur/381332', 10), ('heur/346300', 11), ('heur/379318', 11), ('heur/379307', 17), ('heur/346296', 21), ('heur/330959', 23), ('heur/346301', 25), ('heur/346294', 26), ('heur/379313', 40), ('heur/381331', 78)],
105: [('heur/330959', 1), ('heur/386350', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409478', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/381328', 12), ('heur/346301', 21), ('heur/346300', 23), ('heur/346295', 59), ('heur/346302', 59), ('heur/346298', 96)],
106: [('heur/457562', 5), ('heur/346303', 6), ('heur/346296', 10), ('heur/346292', 106), ('heur/346291', 173)],
107: [('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346303', 5), ('heur/379314', 16), ('heur/346291', 45), ('heur/346293', 66), ('heur/346292', 141)],
108: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386344', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/421526', 165)],
109: [('heur/330959', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/471148', 1), ('heur/397453', 293)],
110: [('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 5), ('heur/330959', 9), ('heur/346292', 49)],
111: [('heur/433534', 19), ('heur/346292', 52), ('heur/346299', 229)],
112: [('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 76), ('heur/330959', 103)],
113: [('heur/330959', 1), ('heur/409478', 1), ('heur/379320', 7), ('heur/346300', 21), ('heur/346299', 81), ('heur/381331', 189)],
114: [('heur/330959', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/421509', 12), ('heur/346291', 67)],
115: [('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379308', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381331', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386349', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409476', 1), ('heur/409478', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/381332', 7), ('heur/330959', 11), ('heur/346291', 163)],
116: [('heur/457559', 1), ('heur/471148', 1), ('heur/386351', 9), ('heur/330959', 106), ('heur/346291', 183)],
117: [('heur/379308', 9), ('heur/397418', 11), ('heur/346293', 40), ('heur/389408', 41), ('heur/346302', 59), ('heur/346291', 140)],
118: [('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379305', 5), ('heur/379309', 14), ('heur/346294', 21), ('heur/346291', 41), ('heur/346292', 41), ('heur/330959', 57)],
119: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397419', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457567', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/379308', 184)],
120: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379305', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381324', 1), ('heur/381331', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421524', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433536', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/409455', 10), ('heur/346304', 206)],
121: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379314', 1), ('heur/379317', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383337', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397439', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421529', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457567', 1), ('heur/471148', 1), ('heur/383335', 224)],
122: [('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 15), ('heur/346291', 161)],
123: [('heur/346295', 300)],
124: [('heur/381323', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397445', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409465', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421505', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433541', 1), ('heur/457559', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469575', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346302', 12), ('heur/346294', 13), ('heur/346295', 32), ('heur/346291', 182)],
125: [('heur/386349', 1), ('heur/409473', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346299', 9), ('heur/379306', 14), ('heur/346294', 46), ('heur/346304', 57), ('heur/330959', 66), ('heur/346305', 93)],
126: [('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 5), ('heur/379309', 5), ('heur/346300', 7), ('heur/346295', 9), ('heur/346302', 10), ('heur/346291', 31), ('heur/346294', 47)],
127: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397422', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409472', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421503', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379318', 9), ('heur/330959', 60), ('heur/346298', 112)],
128: [('heur/330959', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421503', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379307', 15), ('heur/346294', 22), ('heur/346298', 123)],
129: [('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/330959', 6), ('heur/346295', 11), ('heur/457560', 13), ('heur/346294', 208)],
130: [('heur/409477', 1), ('heur/409483', 1), ('heur/409485', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379309', 7), ('heur/346293', 12), ('heur/346296', 15), ('heur/346297', 15), ('heur/346295', 16), ('heur/346302', 31), ('heur/346300', 36), ('heur/346291', 51), ('heur/346294', 89)],
131: [('heur/346302', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346295', 5), ('heur/346294', 6), ('heur/346293', 14), ('heur/346291', 143)],
132: [('heur/346291', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379313', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383337', 1), ('heur/383340', 1), ('heur/386343', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/389408', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397442', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409482', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/409462', 5), ('heur/379308', 8), ('heur/346292', 47), ('heur/330959', 145)],
133: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346303', 8), ('heur/346304', 12), ('heur/346302', 14), ('heur/346291', 77)],
134: [('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/445545', 7), ('heur/397442', 13), ('heur/346301', 37), ('heur/346291', 38), ('heur/346303', 146)],
135: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 16), ('heur/330959', 64)],
136: [('heur/469579', 1), ('heur/471148', 1), ('heur/381322', 10), ('heur/379320', 12), ('heur/379316', 14), ('heur/346300', 19), ('heur/386345', 19), ('heur/330959', 38), ('heur/346295', 186)],
137: [('heur/445550', 300)],
138: [('heur/330959', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 7), ('heur/379306', 11), ('heur/346298', 12), ('heur/346296', 17), ('heur/346291', 71), ('heur/346293', 84)],
139: [('heur/330959', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/421497', 1), ('heur/346296', 14), ('heur/346300', 22), ('heur/346302', 51), ('heur/346294', 66), ('heur/346298', 143)],
140: [('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346300', 7), ('heur/346302', 7), ('heur/379304', 9), ('heur/330959', 11), ('heur/346291', 41), ('heur/346293', 92)],
141: [('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379307', 7), ('heur/346291', 12), ('heur/346296', 13), ('heur/346295', 14), ('heur/346292', 16), ('heur/330959', 135)],
142: [('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/397429', 5), ('heur/379307', 11), ('heur/397428', 15), ('heur/383340', 18), ('heur/397423', 18), ('heur/397427', 18), ('heur/346297', 40), ('heur/330959', 56), ('heur/346298', 103)],
143: [('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/379314', 6), ('heur/346297', 287)],
144: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397427', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421504', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/409475', 8), ('heur/330959', 173)],
145: [('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 16), ('heur/379307', 19), ('heur/346295', 37), ('heur/346301', 219)],
146: [('heur/330959', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397444', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409486', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/433530', 9), ('heur/433534', 11), ('heur/346291', 41), ('heur/379313', 150)],
147: [('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 11), ('heur/346297', 13), ('heur/346303', 15), ('heur/330959', 50), ('heur/346291', 72), ('heur/346294', 92)],
148: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/383340', 1), ('heur/386345', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397422', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397442', 1), ('heur/397449', 1), ('heur/397452', 1), ('heur/409456', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421498', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445543', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/471148', 1), ('heur/469577', 236)],
149: [('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346295', 13), ('heur/330959', 41), ('heur/346291', 188)],
150: [('heur/330959', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397427', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397445', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409469', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/421494', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421521', 1), ('heur/421526', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/379312', 222)],
151: [('heur/330959', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409468', 1), ('heur/409477', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421509', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/397435', 247)],
152: [('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 6), ('heur/346291', 19), ('heur/330959', 51), ('heur/346293', 147)],
153: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/397422', 6), ('heur/346291', 80)],
154: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383337', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397446', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/469568', 14), ('heur/330959', 130)],
155: [('heur/421514', 13), ('heur/346294', 287)],
156: [('heur/330959', 1), ('heur/386348', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409474', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457558', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 6), ('heur/379305', 14), ('heur/379304', 20), ('heur/346296', 21), ('heur/346297', 57), ('heur/346301', 106)],
157: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421504', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421517', 1), ('heur/421523', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/469566', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/471148', 1), ('heur/409476', 5), ('heur/397450', 22), ('heur/346291', 41), ('heur/397431', 121)],
158: [('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/397437', 5), ('heur/330959', 7), ('heur/346291', 8), ('heur/346292', 61), ('heur/346302', 133)],
159: [('heur/397437', 300)],
160: [('heur/330959', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346297', 5), ('heur/346291', 17), ('heur/346294', 48), ('heur/346293', 191)],
161: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/379307', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381334', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397453', 1), ('heur/409463', 1), ('heur/409475', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/346295', 7), ('heur/346291', 221)],
162: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/386345', 1), ('heur/386350', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397441', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409458', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/421504', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421526', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/445546', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457559', 1), ('heur/457565', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397442', 6), ('heur/379310', 14), ('heur/346295', 20), ('heur/346302', 198)],
163: [('heur/330959', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433539', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346303', 252)],
164: [('heur/330959', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/469569', 290)],
165: [('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/330959', 23), ('heur/346296', 41), ('heur/346292', 42), ('heur/346291', 115)],
166: [('heur/421520', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346291', 52), ('heur/346294', 59), ('heur/346292', 64), ('heur/346296', 85)],
167: [('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346293', 8), ('heur/346294', 8), ('heur/346295', 14), ('heur/330959', 71), ('heur/346291', 75)],
168: [('heur/330959', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 112), ('heur/346294', 183)],
169: [('heur/346292', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/379308', 1), ('heur/381332', 1), ('heur/397421', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/421507', 1), ('heur/421514', 1), ('heur/457557', 1), ('heur/457567', 1), ('heur/346295', 5), ('heur/397423', 14), ('heur/346291', 49), ('heur/346296', 218)],
170: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381326', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/433529', 1), ('heur/433541', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346296', 11), ('heur/330959', 17), ('heur/346293', 30), ('heur/379314', 132)],
171: [('heur/421511', 13), ('heur/379307', 287)],
172: [('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 21), ('heur/346292', 71), ('heur/379314', 151)],
173: [('heur/346291', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/471148', 1), ('heur/330959', 23), ('heur/346296', 31), ('heur/346295', 33), ('heur/346292', 122)],
174: [('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397430', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/409455', 1), ('heur/409462', 1), ('heur/409465', 1), ('heur/409475', 1), ('heur/409478', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421521', 1), ('heur/421526', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346291', 11), ('heur/346296', 17), ('heur/346294', 229)],
175: [('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346296', 7), ('heur/346292', 11), ('heur/330959', 68), ('heur/346291', 80)],
176: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409460', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/445548', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457558', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346303', 7), ('heur/379306', 7), ('heur/346291', 62), ('heur/346294', 115)],
177: [('heur/409477', 5), ('heur/397451', 8), ('heur/457559', 8), ('heur/383336', 279)],
178: [('heur/330959', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/409463', 8), ('heur/346292', 61), ('heur/397431', 185)],
179: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433535', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469577', 1), ('heur/330959', 146)],
180: [('heur/330959', 1), ('heur/409472', 1), ('heur/409479', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/421509', 5), ('heur/346296', 7), ('heur/379317', 13), ('heur/346291', 102), ('heur/409477', 155)],
181: [('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346292', 11), ('heur/346291', 13), ('heur/330959', 94)],
182: [('heur/421523', 12), ('heur/346291', 288)],
183: [('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397439', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/346294', 7), ('heur/346296', 11), ('heur/346295', 12), ('heur/346293', 161)],
184: [('heur/397435', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397451', 1), ('heur/397454', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409475', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421508', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469578', 1), ('heur/346293', 7), ('heur/379315', 245)],
185: [('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379318', 6), ('heur/346292', 7), ('heur/346303', 34), ('heur/346293', 35), ('heur/346294', 59)],
186: [('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/346302', 5), ('heur/346301', 14), ('heur/346293', 38), ('heur/346291', 51), ('heur/346294', 102)],
187: [('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421517', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/346293', 8), ('heur/346291', 23), ('heur/346294', 140)],
188: [('heur/330959', 1), ('heur/379306', 5), ('heur/381322', 5), ('heur/381325', 5), ('heur/346299', 6), ('heur/346302', 6), ('heur/346300', 8), ('heur/346301', 8), ('heur/346304', 8), ('heur/379304', 17), ('heur/346297', 27), ('heur/346293', 204)],
189: [('heur/346292', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346291', 7), ('heur/330959', 103)],
190: [('heur/469567', 5), ('heur/346291', 41), ('heur/469574', 254)],
191: [('heur/346295', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457568', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346292', 5), ('heur/397437', 5), ('heur/346297', 19), ('heur/346293', 21), ('heur/330959', 34), ('heur/346291', 77)],
192: [('heur/330959', 1), ('heur/346293', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445544', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346292', 6), ('heur/346295', 6), ('heur/346300', 30), ('heur/346291', 86)],
193: [('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 6), ('heur/346297', 11), ('heur/346292', 20), ('heur/346291', 22), ('heur/330959', 29)],
194: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 5), ('heur/346297', 9), ('heur/330959', 132)],
195: [('heur/346291', 300)],
196: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397449', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421514', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433536', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 11), ('heur/346295', 200)],
197: [('heur/330959', 1), ('heur/379304', 1), ('heur/389351', 1), ('heur/397428', 1), ('heur/409483', 1), ('heur/421498', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/469577', 1), ('heur/409456', 5), ('heur/397447', 14), ('heur/397418', 16), ('heur/397437', 16), ('heur/397426', 17), ('heur/389408', 31), ('heur/397435', 92), ('heur/346295', 100)],
198: [('heur/330959', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346297', 7), ('heur/346293', 23), ('heur/346295', 47), ('heur/346291', 92)],
199: [('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346301', 5), ('heur/346296', 17), ('heur/330959', 32), ('heur/346292', 68)],
200: [('heur/409469', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445549', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346298', 7), ('heur/346302', 9), ('heur/346294', 260)],
201: [('heur/346297', 1), ('heur/346305', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379318', 1), ('heur/397429', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397452', 1), ('heur/409468', 1), ('heur/409480', 1), ('heur/421502', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469574', 1), ('heur/346293', 279)],
202: [('heur/330959', 1), ('heur/346298', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379313', 1), ('heur/381324', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/386343', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/471148', 1), ('heur/397433', 9), ('heur/381328', 228)],
203: [('heur/397421', 7), ('heur/346293', 43), ('heur/346291', 250)],
204: [('heur/346292', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383333', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/389351', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409465', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421499', 1), ('heur/421504', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346305', 5), ('heur/346291', 11), ('heur/379305', 202)],
205: [('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 5), ('heur/346294', 10), ('heur/346302', 12), ('heur/346295', 115), ('heur/346293', 154)],
206: [('heur/346302', 1), ('heur/346305', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346304', 5), ('heur/421503', 7), ('heur/397421', 11), ('heur/346293', 17), ('heur/330959', 19), ('heur/346291', 41), ('heur/346303', 50)],
207: [('heur/409469', 1), ('heur/409471', 1), ('heur/421497', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/381327', 6), ('heur/397454', 10), ('heur/379312', 13), ('heur/346296', 21), ('heur/397416', 24), ('heur/330959', 25), ('heur/379314', 63), ('heur/397450', 130)],
208: [('heur/409466', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346296', 13), ('heur/346297', 13), ('heur/379313', 65), ('heur/346300', 75), ('heur/330959', 120)],
209: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381326', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397441', 1), ('heur/397445', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421498', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/386348', 220)],
210: [('heur/346292', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/381321', 1), ('heur/383337', 1), ('heur/383341', 1), ('heur/397418', 1), ('heur/397431', 1), ('heur/409461', 1), ('heur/409477', 1), ('heur/409482', 1), ('heur/445549', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/346295', 42), ('heur/346291', 236)],
211: [('heur/409479', 300)],
212: [('heur/409466', 300)],
213: [('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 6), ('heur/346291', 288)],
214: [('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/433538', 6), ('heur/346300', 12), ('heur/397441', 12), ('heur/330959', 29), ('heur/346295', 30), ('heur/379314', 136)],
215: [('heur/409469', 1), ('heur/409476', 1), ('heur/409482', 1), ('heur/421492', 1), ('heur/421501', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346300', 5), ('heur/330959', 12), ('heur/346295', 13), ('heur/346298', 22), ('heur/346305', 23), ('heur/379314', 63), ('heur/346291', 124)],
216: [('heur/330959', 1), ('heur/421496', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/457559', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/471148', 1), ('heur/379306', 13), ('heur/379309', 18), ('heur/346294', 51), ('heur/381331', 59), ('heur/397435', 142)],
217: [('heur/421496', 7), ('heur/346302', 34), ('heur/346304', 55), ('heur/346292', 204)],
218: [('heur/330959', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/381331', 1), ('heur/389408', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409473', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/397421', 19), ('heur/346291', 245)],
219: [('heur/346291', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/397417', 1), ('heur/397442', 1), ('heur/397451', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409477', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421519', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/346294', 13), ('heur/346293', 17), ('heur/346295', 24), ('heur/346296', 182)],
220: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381324', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383338', 1), ('heur/386343', 1), ('heur/386349', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/469576', 14), ('heur/330959', 160)],
221: [('heur/346296', 7), ('heur/346300', 8), ('heur/346301', 11), ('heur/330959', 14), ('heur/346292', 23), ('heur/346291', 31), ('heur/346298', 52), ('heur/346297', 154)],
222: [('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409490', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346292', 5), ('heur/346291', 11), ('heur/330959', 103)],
223: [('heur/330959', 1), ('heur/379319', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421523', 1), ('heur/433529', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/397426', 5), ('heur/386345', 8), ('heur/346291', 226)],
224: [('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 32), ('heur/346304', 55), ('heur/330959', 186)],
225: [('heur/409486', 1), ('heur/409487', 1), ('heur/421496', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/445542', 13), ('heur/346303', 42), ('heur/346294', 68), ('heur/330959', 133)],
226: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379308', 7), ('heur/379313', 182)],
227: [('heur/469576', 1), ('heur/397430', 4), ('heur/386347', 5), ('heur/409466', 7), ('heur/346294', 9), ('heur/397451', 16), ('heur/346300', 77), ('heur/346304', 181)],
228: [('heur/381330', 1), ('heur/409479', 1), ('heur/409488', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421516', 1), ('heur/471148', 1), ('heur/330959', 5), ('heur/346301', 5), ('heur/346300', 10), ('heur/346305', 12), ('heur/346297', 23), ('heur/346292', 26), ('heur/346291', 211)],
229: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/445554', 70)],
230: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379308', 1), ('heur/379313', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/389408', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409468', 1), ('heur/409477', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421503', 1), ('heur/421511', 1), ('heur/421519', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/457558', 1), ('heur/457562', 1), ('heur/469569', 1), ('heur/346305', 258)],
231: [('heur/433536', 9), ('heur/381332', 291)],
232: [('heur/330959', 1), ('heur/346292', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/383335', 1), ('heur/383338', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386346', 1), ('heur/386349', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397446', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409478', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/445549', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346304', 5), ('heur/379309', 6), ('heur/346291', 176)],
233: [('heur/409490', 1), ('heur/409491', 1), ('heur/421501', 1), ('heur/421508', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346302', 5), ('heur/445546', 6), ('heur/346292', 21), ('heur/346295', 22), ('heur/346293', 34), ('heur/330959', 36), ('heur/346291', 138)],
234: [('heur/346291', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379315', 1), ('heur/379320', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397446', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346302', 22), ('heur/346301', 27), ('heur/330959', 44), ('heur/379313', 131)],
235: [('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397419', 11), ('heur/379308', 15), ('heur/330959', 96), ('heur/346300', 118)],
236: [('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 12), ('heur/346291', 84)],
237: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379305', 67)],
238: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346302', 5), ('heur/330959', 9), ('heur/346291', 67)],
239: [('heur/469578', 7), ('heur/457567', 8), ('heur/397422', 285)],
240: [('heur/379318', 300)],
241: [('heur/346291', 1), ('heur/346292', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379309', 1), ('heur/379313', 1), ('heur/381321', 1), ('heur/383339', 1), ('heur/389408', 1), ('heur/397418', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/409461', 1), ('heur/409472', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/421498', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421528', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/457558', 12), ('heur/330959', 246)],
242: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433536', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/346294', 6), ('heur/346291', 153)],
243: [('heur/330959', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/379309', 19), ('heur/346295', 28), ('heur/346291', 128)],
244: [('heur/346291', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457566', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/421518', 6), ('heur/346294', 9), ('heur/346292', 19), ('heur/330959', 120)],
245: [('heur/330959', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/471148', 1), ('heur/397431', 296)],
246: [('heur/457561', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379314', 14), ('heur/346292', 19), ('heur/346295', 120), ('heur/346291', 132)],
247: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/379306', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397445', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409481', 1), ('heur/421499', 1), ('heur/421504', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/457557', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/379307', 235)],
248: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379318', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397442', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409470', 1), ('heur/409480', 1), ('heur/421514', 1), ('heur/421518', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/457562', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346294', 247)],
249: [('heur/330959', 1), ('heur/346291', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/409482', 16), ('heur/346292', 129)],
250: [('heur/330959', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346298', 8), ('heur/433534', 8), ('heur/379308', 14), ('heur/421509', 16), ('heur/346292', 38), ('heur/346295', 39), ('heur/346296', 46), ('heur/379313', 60)],
251: [('heur/379319', 5), ('heur/379306', 20), ('heur/346296', 275)],
252: [('heur/383333', 300)],
253: [('heur/330959', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379316', 5), ('heur/346304', 21), ('heur/346300', 23), ('heur/346291', 178)],
254: [('heur/346291', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397450', 1), ('heur/409461', 1), ('heur/421518', 1), ('heur/421526', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/471148', 1), ('heur/397445', 7), ('heur/330959', 274)],
255: [('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433540', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 10), ('heur/346293', 13), ('heur/346291', 77)],
256: [('heur/346292', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379304', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381324', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409476', 1), ('heur/409479', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421508', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/330959', 25), ('heur/346304', 27), ('heur/346291', 135)],
257: [('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/433534', 5), ('heur/397437', 6), ('heur/379313', 10), ('heur/346292', 11), ('heur/330959', 130), ('heur/346293', 130)],
258: [('heur/346298', 1), ('heur/346305', 1), ('heur/379308', 1), ('heur/379313', 1), ('heur/381332', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457557', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/346296', 6), ('heur/346294', 278)],
259: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409470', 1), ('heur/409477', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421499', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433540', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/346297', 7), ('heur/346293', 9), ('heur/379315', 15), ('heur/379313', 201)],
260: [('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409469', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346298', 5), ('heur/330959', 6), ('heur/379312', 9), ('heur/346294', 21), ('heur/346292', 52), ('heur/346291', 93)],
261: [('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 8), ('heur/330959', 12), ('heur/346297', 18), ('heur/346296', 21), ('heur/346300', 21), ('heur/346291', 74)],
262: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457561', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/346293', 188)],
263: [('heur/330959', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/397416', 274)],
264: [('heur/330959', 1), ('heur/379318', 5), ('heur/397422', 11), ('heur/409474', 13), ('heur/389408', 120), ('heur/346304', 150)],
265: [('heur/445545', 1), ('heur/445546', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/346294', 35), ('heur/346291', 252)],
266: [('heur/409457', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397422', 6), ('heur/381333', 10), ('heur/346302', 28), ('heur/379314', 173)],
267: [('heur/346297', 5), ('heur/330959', 57), ('heur/346291', 101), ('heur/346296', 137)],
268: [('heur/346291', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 13), ('heur/346292', 62)],
269: [('heur/346291', 1), ('heur/346293', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383341', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397449', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445549', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 8), ('heur/346295', 10), ('heur/346296', 11), ('heur/346292', 156)],
270: [('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/433541', 1), ('heur/445543', 1), ('heur/445546', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346291', 5), ('heur/346293', 11), ('heur/346294', 17), ('heur/330959', 44), ('heur/346292', 55)],
271: [('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 5), ('heur/346295', 6), ('heur/330959', 33), ('heur/346291', 152)],
272: [('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397442', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346302', 23), ('heur/379313', 23), ('heur/346291', 32), ('heur/346294', 83)],
273: [('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379320', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389352', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433539', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469571', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346304', 11), ('heur/346300', 14), ('heur/346292', 138)],
274: [('heur/330959', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/383337', 1), ('heur/389350', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397450', 1), ('heur/409455', 1), ('heur/409461', 1), ('heur/409478', 1), ('heur/409482', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421496', 1), ('heur/421511', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346292', 11), ('heur/346291', 228)],
275: [('heur/346299', 300)],
276: [('heur/346292', 1), ('heur/346294', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379307', 1), ('heur/379314', 1), ('heur/381321', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386348', 1), ('heur/397423', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409480', 1), ('heur/421493', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421520', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/457559', 1), ('heur/469579', 1), ('heur/381332', 9), ('heur/346293', 12), ('heur/379308', 18), ('heur/346296', 218)],
277: [('heur/346291', 1), ('heur/346292', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/346295', 8), ('heur/330959', 157)],
278: [('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/330959', 25), ('heur/346296', 40), ('heur/346302', 49), ('heur/346291', 164)],
279: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433540', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457569', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 10), ('heur/330959', 80)],
280: [('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 7), ('heur/346295', 7), ('heur/330959', 9), ('heur/346292', 11), ('heur/346291', 257)],
281: [('heur/330959', 1), ('heur/409468', 1), ('heur/409475', 1), ('heur/409483', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346298', 20), ('heur/346292', 65), ('heur/346291', 72), ('heur/346303', 103)],
282: [('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346300', 10), ('heur/330959', 24), ('heur/346291', 90)],
283: [('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 25), ('heur/346292', 41), ('heur/346291', 72)],
284: [('heur/397430', 300)],
285: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346291', 106)],
286: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 5), ('heur/346297', 6), ('heur/379304', 7), ('heur/397439', 8), ('heur/379306', 9), ('heur/397423', 9), ('heur/409465', 10), ('heur/379318', 12), ('heur/397436', 13), ('heur/379308', 14), ('heur/445546', 15), ('heur/346295', 58), ('heur/346303', 65), ('heur/346298', 67)],
287: [('heur/421506', 300)],
288: [('heur/445546', 300)],
289: [('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346301', 5), ('heur/346295', 10), ('heur/346292', 41), ('heur/346293', 43), ('heur/330959', 69)],
290: [('heur/379312', 300)],
291: [('heur/330959', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 21), ('heur/346302', 42), ('heur/346294', 160)],
292: [('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346295', 15), ('heur/346291', 16), ('heur/330959', 45), ('heur/346293', 46), ('heur/346302', 160)],
293: [('heur/386350', 1), ('heur/409466', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346295', 8), ('heur/346292', 13), ('heur/330959', 18), ('heur/346291', 38), ('heur/346294', 123)],
294: [('heur/330959', 1), ('heur/346291', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409476', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/409460', 11), ('heur/346292', 170)],
295: [('heur/346304', 300)],
296: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409469', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/433534', 148)],
297: [('heur/330959', 1), ('heur/457565', 7), ('heur/346292', 24), ('heur/346301', 71), ('heur/346300', 75), ('heur/457559', 122)],
298: [('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346291', 73), ('heur/346292', 222)],
299: [('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 11), ('heur/346294', 15), ('heur/330959', 40), ('heur/346291', 49)],
300: [('heur/421507', 300)],
301: [('heur/471148', 1), ('heur/330959', 11), ('heur/346297', 18), ('heur/379304', 20), ('heur/346302', 88), ('heur/346296', 162)],
302: [('heur/330959', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/409465', 5), ('heur/346294', 8), ('heur/346297', 8), ('heur/346302', 11), ('heur/346298', 12), ('heur/346293', 33), ('heur/379314', 47), ('heur/346291', 159)],
303: [('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 28), ('heur/346296', 28), ('heur/346291', 61), ('heur/346294', 170)],
304: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/389408', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/421507', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421528', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445544', 1), ('heur/457559', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346296', 261)],
305: [('heur/381334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445544', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/346291', 6), ('heur/346292', 6), ('heur/346296', 169)],
306: [('heur/330959', 1), ('heur/379306', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409488', 1), ('heur/421495', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/379313', 5), ('heur/346305', 6), ('heur/379304', 6), ('heur/379309', 6), ('heur/379312', 8), ('heur/397438', 9), ('heur/346300', 15), ('heur/379310', 19), ('heur/346295', 23), ('heur/346294', 100)],
307: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 87)],
308: [('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346294', 6), ('heur/346300', 31), ('heur/346298', 219)],
309: [('heur/346292', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386349', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409460', 1), ('heur/409468', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421501', 1), ('heur/421506', 1), ('heur/421510', 1), ('heur/421516', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/445542', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/469567', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/346291', 201)],
310: [('heur/330959', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379320', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409478', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346291', 11), ('heur/346300', 155)],
311: [('heur/330959', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457562', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346297', 5), ('heur/379305', 5), ('heur/379306', 5), ('heur/346295', 6), ('heur/346294', 11), ('heur/346291', 24), ('heur/346292', 67), ('heur/346298', 81)],
312: [('heur/397441', 300)],
313: [('heur/330959', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/379305', 7), ('heur/397424', 15), ('heur/346291', 21), ('heur/397416', 233)],
314: [('heur/346294', 1), ('heur/346297', 1), ('heur/346302', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379319', 1), ('heur/383340', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397451', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/421498', 1), ('heur/421509', 1), ('heur/421514', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/457569', 1), ('heur/421531', 5), ('heur/346301', 8), ('heur/381332', 9), ('heur/379306', 10), ('heur/346295', 19), ('heur/379313', 222)],
315: [('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 5), ('heur/330959', 7), ('heur/346292', 24), ('heur/346298', 261)],
316: [('heur/330959', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397445', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346295', 9), ('heur/346298', 39), ('heur/346293', 49), ('heur/346294', 88)],
317: [('heur/469576', 1), ('heur/471148', 1), ('heur/379305', 5), ('heur/379312', 18), ('heur/381331', 25), ('heur/397416', 28), ('heur/346292', 64), ('heur/330959', 158)],
318: [('heur/330959', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445548', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346302', 216)],
319: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 18), ('heur/346291', 125)],
320: [('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379308', 1), ('heur/379314', 1), ('heur/381321', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/409479', 1), ('heur/421512', 1), ('heur/421521', 1), ('heur/457559', 1), ('heur/469569', 1), ('heur/471148', 1), ('heur/381326', 5), ('heur/379307', 275)],
321: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/421494', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/397423', 7), ('heur/379308', 228)],
322: [('heur/346292', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379313', 1), ('heur/389408', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409469', 1), ('heur/409472', 1), ('heur/409477', 1), ('heur/421505', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/330959', 55), ('heur/346291', 224)],
323: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397441', 1), ('heur/397447', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421492', 1), ('heur/421500', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/457561', 1), ('heur/469573', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346297', 6), ('heur/346291', 201)],
324: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383336', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/389351', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421521', 1), ('heur/421531', 1), ('heur/457559', 1), ('heur/469579', 1), ('heur/379308', 239)],
325: [('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397442', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/346291', 41), ('heur/346293', 90), ('heur/346296', 94)],
326: [('heur/471148', 1), ('heur/381332', 5), ('heur/381331', 7), ('heur/346295', 8), ('heur/346291', 11), ('heur/346301', 11), ('heur/381326', 17), ('heur/346296', 101), ('heur/379313', 139)],
327: [('heur/346293', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/457559', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 8), ('heur/346302', 12), ('heur/346292', 16), ('heur/346295', 26), ('heur/346291', 119)],
328: [('heur/421496', 6), ('heur/409455', 7), ('heur/469575', 287)],
329: [('heur/381327', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397418', 1), ('heur/397421', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409466', 1), ('heur/409470', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/457565', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346291', 11), ('heur/346293', 19), ('heur/346305', 56), ('heur/330959', 168)],
330: [('heur/346291', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 6), ('heur/346292', 29), ('heur/346295', 37), ('heur/346294', 74)],
331: [('heur/421496', 1), ('heur/421498', 1), ('heur/421505', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/346295', 5), ('heur/381327', 5), ('heur/381334', 5), ('heur/421520', 5), ('heur/346294', 8), ('heur/346291', 11), ('heur/346292', 11), ('heur/346298', 12), ('heur/346296', 35), ('heur/346293', 67), ('heur/379314', 82)],
332: [('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379319', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/346293', 10), ('heur/346297', 11), ('heur/346291', 105)],
333: [('heur/409470', 1), ('heur/409473', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 5), ('heur/346301', 6), ('heur/346293', 8), ('heur/346302', 8), ('heur/346295', 10), ('heur/346296', 11), ('heur/346291', 161)],
334: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/330959', 11), ('heur/346291', 52)],
335: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/381326', 1), ('heur/381331', 1), ('heur/383340', 1), ('heur/386347', 1), ('heur/389408', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397442', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445549', 1), ('heur/457559', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 22), ('heur/330959', 226)],
336: [('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379320', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397429', 1), ('heur/397452', 1), ('heur/409464', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/346291', 6), ('heur/346294', 247)],
337: [('heur/330959', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346291', 9), ('heur/379308', 19), ('heur/346296', 26), ('heur/379313', 27), ('heur/346294', 51)],
338: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346291', 106)],
339: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397442', 6), ('heur/346295', 14), ('heur/346291', 134)],
340: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421499', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421531', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469579', 1), ('heur/346293', 228)],
341: [('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346292', 27), ('heur/346291', 73), ('heur/330959', 86)],
342: [('heur/346294', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/330959', 11), ('heur/346293', 15), ('heur/346297', 18), ('heur/346303', 36), ('heur/346300', 40), ('heur/346292', 41), ('heur/346304', 54), ('heur/346295', 70)],
343: [('heur/330959', 1), ('heur/471148', 1), ('heur/346292', 111), ('heur/346304', 187)],
344: [('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 6), ('heur/330959', 7), ('heur/346291', 21), ('heur/346304', 46), ('heur/346296', 48), ('heur/346293', 64), ('heur/346295', 77)],
345: [('heur/421495', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/397422', 5), ('heur/346292', 6), ('heur/346295', 6), ('heur/381328', 7), ('heur/346301', 8), ('heur/346297', 9), ('heur/346296', 11), ('heur/346303', 18), ('heur/346299', 21), ('heur/330959', 28), ('heur/346294', 176)],
346: [('heur/346291', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386348', 1), ('heur/389350', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397441', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409477', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/457556', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/346296', 5), ('heur/346294', 11), ('heur/346293', 16), ('heur/379308', 18), ('heur/330959', 147)],
347: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379308', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383340', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409466', 1), ('heur/409478', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421518', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/457559', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/433541', 254)],
348: [('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/381323', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381334', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386351', 1), ('heur/397416', 1), ('heur/397421', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/421493', 1), ('heur/421498', 1), ('heur/421505', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421521', 1), ('heur/421528', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379305', 17), ('heur/346292', 25), ('heur/346291', 202)],
349: [('heur/330959', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383338', 1), ('heur/383340', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/421495', 1), ('heur/421498', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/433537', 221)],
350: [('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 9), ('heur/330959', 10), ('heur/346291', 11), ('heur/381331', 89)],
351: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379305', 7), ('heur/330959', 11), ('heur/346291', 99)],
352: [('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/386348', 5), ('heur/330959', 15), ('heur/409463', 20), ('heur/346291', 41), ('heur/379313', 178)],
353: [('heur/381328', 300)],
354: [('heur/330959', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 13), ('heur/346303', 95)],
355: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346305', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379320', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397445', 1), ('heur/397451', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469579', 1), ('heur/379307', 8), ('heur/346293', 221)],
356: [('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379310', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397424', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409468', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421505', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/445548', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346291', 6), ('heur/330959', 198)],
357: [('heur/330959', 1), ('heur/346293', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/381331', 1), ('heur/383337', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433532', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397452', 10), ('heur/409457', 13), ('heur/379313', 33), ('heur/346299', 135)],
358: [('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346303', 7), ('heur/346295', 15), ('heur/346293', 32), ('heur/346294', 35), ('heur/346291', 91)],
359: [('heur/421527', 5), ('heur/433536', 5), ('heur/433534', 8), ('heur/469575', 10), ('heur/379314', 272)],
360: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379309', 1), ('heur/379313', 1), ('heur/381323', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386349', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409483', 1), ('heur/421498', 1), ('heur/421504', 1), ('heur/421523', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/457559', 1), ('heur/471148', 1), ('heur/381329', 9), ('heur/330959', 247)],
361: [('heur/346292', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/379315', 5), ('heur/330959', 9), ('heur/346291', 97)],
362: [('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346298', 5), ('heur/330959', 6), ('heur/346297', 6), ('heur/346293', 8), ('heur/346291', 14), ('heur/346300', 28), ('heur/346296', 225)],
363: [('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 9), ('heur/346297', 11), ('heur/346291', 17), ('heur/346294', 92)],
364: [('heur/330959', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397416', 66), ('heur/346302', 92)],
365: [('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 8), ('heur/346296', 11), ('heur/346294', 22), ('heur/330959', 41)],
366: [('heur/346291', 1), ('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/379305', 1), ('heur/379315', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/397419', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/397449', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/421493', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433536', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445549', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469577', 1), ('heur/346293', 15), ('heur/379308', 16), ('heur/379313', 215)],
367: [('heur/330959', 8), ('heur/389408', 292)],
368: [('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379308', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397426', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421510', 1), ('heur/433531', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/383337', 6), ('heur/346291', 108), ('heur/346294', 153)],
369: [('heur/330959', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 13), ('heur/346293', 29), ('heur/346291', 63)],
370: [('heur/346291', 1), ('heur/346292', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/379307', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/397435', 1), ('heur/397453', 1), ('heur/409480', 1), ('heur/421500', 1), ('heur/421526', 1), ('heur/386347', 282)],
371: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/379304', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346303', 6), ('heur/346295', 7), ('heur/379308', 10), ('heur/379305', 17), ('heur/346304', 41), ('heur/346291', 60)],
372: [('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445544', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/330959', 91), ('heur/346291', 139)],
373: [('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 13), ('heur/346296', 238)],
374: [('heur/346291', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/381331', 1), ('heur/389408', 1), ('heur/397424', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409477', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421521', 1), ('heur/421528', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433539', 1), ('heur/457559', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397416', 10), ('heur/346293', 11), ('heur/379308', 13), ('heur/346294', 92), ('heur/409461', 127)],
375: [('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397425', 1), ('heur/397427', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397443', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409468', 1), ('heur/409472', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421500', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/330959', 22), ('heur/346294', 36), ('heur/346292', 148)],
376: [('heur/409471', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409480', 1), ('heur/409484', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346297', 9), ('heur/330959', 10), ('heur/346292', 13), ('heur/346293', 34), ('heur/346295', 68), ('heur/346298', 132)],
377: [('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/346292', 7), ('heur/346291', 11), ('heur/330959', 12), ('heur/346302', 40), ('heur/346298', 79), ('heur/346301', 128)],
378: [('heur/346293', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381325', 1), ('heur/381328', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386348', 1), ('heur/389350', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397421', 1), ('heur/397430', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397442', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/421493', 1), ('heur/421501', 1), ('heur/421503', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421526', 1), ('heur/421529', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/379312', 6), ('heur/346296', 21), ('heur/346292', 196)],
379: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346305', 1), ('heur/379308', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/381326', 1), ('heur/397423', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/409466', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/433539', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/397452', 6), ('heur/346291', 269)],
380: [('heur/346292', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381326', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386344', 1), ('heur/386348', 1), ('heur/397416', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409466', 1), ('heur/409477', 1), ('heur/409480', 1), ('heur/421511', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421525', 1), ('heur/421531', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469579', 1), ('heur/346294', 12), ('heur/346291', 41), ('heur/346293', 189)],
381: [('heur/389350', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/346291', 6), ('heur/346300', 10), ('heur/381330', 10), ('heur/346304', 11), ('heur/379314', 14), ('heur/346302', 17), ('heur/330959', 20), ('heur/346298', 22), ('heur/346293', 187)],
382: [('heur/445542', 1), ('heur/445543', 1), ('heur/445545', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346298', 12), ('heur/346291', 51), ('heur/330959', 72), ('heur/381331', 142)],
383: [('heur/346292', 1), ('heur/346299', 1), ('heur/346304', 1), ('heur/379313', 1), ('heur/381331', 1), ('heur/397419', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409461', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409472', 1), ('heur/409477', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/397416', 56), ('heur/346291', 206)],
384: [('heur/421521', 11), ('heur/346302', 29), ('heur/346300', 56), ('heur/346292', 204)],
385: [('heur/330959', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346294', 6), ('heur/346295', 18), ('heur/346301', 31), ('heur/346302', 43), ('heur/346292', 117)],
386: [('heur/397416', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397445', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 25), ('heur/346293', 185)],
387: [('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433541', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346302', 5), ('heur/379316', 5), ('heur/330959', 13), ('heur/346291', 19), ('heur/346294', 48), ('heur/346298', 138)],
388: [('heur/346292', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 8), ('heur/330959', 21), ('heur/346291', 117)],
389: [('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421509', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/397438', 7), ('heur/346291', 82), ('heur/346300', 146)],
390: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381325', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383336', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397441', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421499', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346297', 5), ('heur/379308', 6), ('heur/346294', 29), ('heur/346296', 61), ('heur/346295', 105)],
391: [('heur/469579', 300)],
392: [('heur/346291', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379315', 1), ('heur/379319', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383335', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/389351', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397421', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397441', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421513', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/445546', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457562', 1), ('heur/457565', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346292', 186)],
393: [('heur/330959', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433536', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346299', 9), ('heur/346291', 131)],
394: [('heur/346294', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379319', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/397422', 1), ('heur/397424', 1), ('heur/397436', 1), ('heur/397442', 1), ('heur/409462', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409477', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421498', 1), ('heur/421500', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421531', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457564', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/346293', 248)],
395: [('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 11), ('heur/330959', 25), ('heur/346304', 56), ('heur/346302', 162)],
396: [('heur/457561', 5), ('heur/346304', 6), ('heur/386345', 6), ('heur/457563', 283)],
397: [('heur/469572', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/379314', 6), ('heur/346301', 22), ('heur/346300', 266)],
398: [('heur/330959', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346298', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397439', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421525', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346300', 6), ('heur/346297', 10), ('heur/346296', 14), ('heur/346299', 19), ('heur/346291', 141)],
399: [('heur/346300', 1), ('heur/397438', 1), ('heur/421525', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/445552', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/379308', 6), ('heur/379310', 7), ('heur/379309', 9), ('heur/379304', 20), ('heur/346295', 22), ('heur/379314', 24), ('heur/346301', 31), ('heur/346302', 37), ('heur/346293', 52), ('heur/330959', 77)],
400: [('heur/386348', 1), ('heur/389351', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421498', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/379304', 6), ('heur/379315', 6), ('heur/381334', 7), ('heur/346292', 8), ('heur/346296', 11), ('heur/379306', 12), ('heur/346297', 19), ('heur/346295', 117)],
401: [('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/381323', 1), ('heur/386344', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409466', 1), ('heur/421494', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421531', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/469568', 1), ('heur/469579', 1), ('heur/397423', 279)],
402: [('heur/346299', 1), ('heur/346300', 1), ('heur/346304', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397424', 1), ('heur/397447', 1), ('heur/397450', 1), ('heur/409463', 1), ('heur/409482', 1), ('heur/421499', 1), ('heur/421509', 1), ('heur/421524', 1), ('heur/421528', 1), ('heur/471148', 1), ('heur/409465', 11), ('heur/397435', 272)],
403: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379314', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409457', 1), ('heur/409461', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409477', 1), ('heur/421500', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/457559', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469577', 1), ('heur/471148', 1), ('heur/346294', 10), ('heur/379313', 13), ('heur/330959', 199)],
404: [('heur/445542', 300)],
405: [('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/381321', 1), ('heur/381326', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386345', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397423', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397446', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421527', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 12), ('heur/346292', 177)],
406: [('heur/330959', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397420', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421518', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/457559', 1), ('heur/457566', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/379309', 5), ('heur/346291', 11), ('heur/346292', 11), ('heur/346293', 134)],
407: [('heur/330959', 1), ('heur/379304', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346296', 10), ('heur/346295', 13), ('heur/346292', 21), ('heur/346291', 40)],
408: [('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/389350', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/330959', 5), ('heur/346298', 8), ('heur/346293', 14), ('heur/346291', 35), ('heur/346292', 62)],
409: [('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381327', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397444', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409466', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421494', 1), ('heur/421500', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/457559', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 7), ('heur/346294', 8), ('heur/346293', 174)],
410: [('heur/469579', 1), ('heur/471148', 1), ('heur/421498', 6), ('heur/346300', 20), ('heur/346296', 50), ('heur/346302', 77), ('heur/346298', 145)],
411: [('heur/330959', 1), ('heur/381331', 1), ('heur/381334', 1), ('heur/386345', 1), ('heur/389408', 1), ('heur/409455', 1), ('heur/433536', 1), ('heur/457559', 1), ('heur/469566', 1), ('heur/471148', 1), ('heur/379304', 6), ('heur/346305', 11), ('heur/346298', 24), ('heur/346292', 249)],
412: [('heur/346291', 1), ('heur/346292', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381325', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397433', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397445', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421522', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/379305', 5), ('heur/346294', 13), ('heur/346301', 22), ('heur/330959', 29), ('heur/346293', 111)],
413: [('heur/330959', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346301', 30), ('heur/346291', 74)],
414: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397423', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409469', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/379318', 5), ('heur/379308', 14), ('heur/346291', 221)],
415: [('heur/346292', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346300', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457567', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346291', 7), ('heur/330959', 166)],
416: [('heur/330959', 1), ('heur/397423', 1), ('heur/346294', 5), ('heur/445543', 5), ('heur/383340', 6), ('heur/346299', 8), ('heur/346301', 9), ('heur/379313', 9), ('heur/421520', 12), ('heur/379314', 16), ('heur/346303', 104), ('heur/346296', 124)],
417: [('heur/346291', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379308', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379317', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383337', 1), ('heur/383339', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386345', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397438', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409463', 1), ('heur/409465', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421506', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421515', 1), ('heur/421517', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 8), ('heur/346296', 21), ('heur/346292', 144)],
418: [('heur/330959', 1), ('heur/421500', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433535', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346292', 39), ('heur/346291', 216)],
419: [('heur/346296', 1), ('heur/346298', 1), ('heur/379306', 1), ('heur/379311', 1), ('heur/379319', 1), ('heur/386344', 1), ('heur/409477', 1), ('heur/383336', 293)],
420: [('heur/330959', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421513', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433537', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346296', 60), ('heur/346291', 109)],
421: [('heur/330959', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/379305', 7), ('heur/379306', 7), ('heur/379308', 8), ('heur/346298', 31), ('heur/346301', 62), ('heur/346291', 175)],
422: [('heur/409462', 1), ('heur/409475', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/346291', 21), ('heur/346294', 241)],
423: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/381331', 1), ('heur/386345', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397424', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397438', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/409457', 1), ('heur/409459', 1), ('heur/409461', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409476', 1), ('heur/409482', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/469575', 1), ('heur/471148', 1), ('heur/346294', 13), ('heur/379304', 19), ('heur/330959', 229)],
424: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381324', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/386342', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421493', 1), ('heur/421496', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421515', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457566', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 6), ('heur/346291', 154)],
425: [('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346291', 11), ('heur/330959', 60)],
426: [('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/397436', 5), ('heur/383340', 14), ('heur/330959', 69), ('heur/346291', 135)],
427: [('heur/346291', 1), ('heur/346293', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/383336', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389408', 1), ('heur/397423', 1), ('heur/397426', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397438', 1), ('heur/397442', 1), ('heur/397445', 1), ('heur/397448', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409455', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409466', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/421496', 1), ('heur/421505', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445550', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/346295', 5), ('heur/346294', 14), ('heur/330959', 29), ('heur/346292', 140)],
428: [('heur/330959', 1), ('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/381331', 1), ('heur/389408', 1), ('heur/397416', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/409461', 1), ('heur/409477', 1), ('heur/421509', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/457559', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/346300', 267)],
429: [('heur/330959', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379312', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397420', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397435', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409461', 1), ('heur/409467', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/421493', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421509', 1), ('heur/421512', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457559', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/471148', 1), ('heur/346292', 88), ('heur/346291', 144)],
430: [('heur/330959', 1), ('heur/346296', 6), ('heur/346291', 14), ('heur/346294', 37), ('heur/346295', 71), ('heur/346300', 171)],
431: [('heur/330959', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379318', 1), ('heur/381321', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397450', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/346291', 103)],
432: [('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/346293', 5), ('heur/346300', 13), ('heur/346294', 92), ('heur/346292', 133)],
433: [('heur/330959', 1), ('heur/421503', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433539', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/346293', 6), ('heur/346291', 8), ('heur/346301', 22), ('heur/379313', 66), ('heur/346292', 139)],
434: [('heur/409464', 1), ('heur/409466', 1), ('heur/409468', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421521', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469578', 1), ('heur/471148', 1), ('heur/330959', 14), ('heur/346293', 203)],
435: [('heur/421523', 1), ('heur/445548', 1), ('heur/445551', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/471148', 1), ('heur/330959', 41), ('heur/346296', 51), ('heur/346303', 196)],
436: [('heur/330959', 1), ('heur/346291', 1), ('heur/346294', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346303', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379311', 1), ('heur/379319', 1), ('heur/381325', 1), ('heur/383336', 1), ('heur/383340', 1), ('heur/386347', 1), ('heur/389408', 1), ('heur/397423', 1), ('heur/397431', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397442', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397453', 1), ('heur/409466', 1), ('heur/409478', 1), ('heur/421524', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433538', 1), ('heur/445544', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/421526', 268)],
437: [('heur/346292', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346304', 1), ('heur/379307', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379319', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/397416', 1), ('heur/397421', 1), ('heur/397441', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409457', 1), ('heur/409462', 1), ('heur/409465', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/421498', 1), ('heur/421501', 1), ('heur/421506', 1), ('heur/421509', 1), ('heur/421511', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421526', 1), ('heur/421528', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469574', 1), ('heur/346294', 5), ('heur/346291', 21), ('heur/346293', 214)],
438: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/379308', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379320', 1), ('heur/381325', 1), ('heur/381332', 1), ('heur/383336', 1), ('heur/386344', 1), ('heur/397423', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397449', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/409466', 1), ('heur/421498', 1), ('heur/421511', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/433529', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/457559', 1), ('heur/379307', 267)],
439: [('heur/346292', 1), ('heur/346293', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379313', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381322', 1), ('heur/381324', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383338', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397418', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397429', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397437', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409468', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409479', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/421492', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/433529', 1), ('heur/445542', 1), ('heur/445546', 1), ('heur/445548', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/469566', 1), ('heur/469568', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/471148', 1), ('heur/330959', 10), ('heur/346291', 154)],
440: [('heur/346293', 1), ('heur/346294', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346305', 1), ('heur/379310', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379318', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381331', 1), ('heur/381333', 1), ('heur/383334', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386350', 1), ('heur/389350', 1), ('heur/389352', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397419', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397432', 1), ('heur/397434', 1), ('heur/397436', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409456', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409464', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409473', 1), ('heur/409475', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409480', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421504', 1), ('heur/421508', 1), ('heur/421510', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421518', 1), ('heur/421520', 1), ('heur/421523', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/433530', 1), ('heur/433532', 1), ('heur/433538', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457562', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469575', 1), ('heur/469577', 1), ('heur/469579', 1), ('heur/379309', 7), ('heur/346292', 11), ('heur/346291', 45), ('heur/346300', 114)],
441: [('heur/346292', 1), ('heur/346296', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/381329', 1), ('heur/386343', 1), ('heur/386345', 1), ('heur/409460', 1), ('heur/409477', 1), ('heur/409481', 1), ('heur/421516', 1), ('heur/421518', 1), ('heur/421521', 1), ('heur/421528', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/457559', 1), ('heur/457564', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469569', 1), ('heur/469574', 1), ('heur/421520', 18), ('heur/346291', 246)],
442: [('heur/346291', 1), ('heur/346292', 1), ('heur/346293', 1), ('heur/346294', 1), ('heur/346295', 1), ('heur/346296', 1), ('heur/346297', 1), ('heur/346298', 1), ('heur/346299', 1), ('heur/346300', 1), ('heur/346301', 1), ('heur/346302', 1), ('heur/346303', 1), ('heur/346304', 1), ('heur/346305', 1), ('heur/379304', 1), ('heur/379305', 1), ('heur/379306', 1), ('heur/379307', 1), ('heur/379308', 1), ('heur/379309', 1), ('heur/379310', 1), ('heur/379311', 1), ('heur/379312', 1), ('heur/379313', 1), ('heur/379314', 1), ('heur/379315', 1), ('heur/379316', 1), ('heur/379317', 1), ('heur/379318', 1), ('heur/379319', 1), ('heur/379320', 1), ('heur/381321', 1), ('heur/381322', 1), ('heur/381323', 1), ('heur/381324', 1), ('heur/381325', 1), ('heur/381326', 1), ('heur/381327', 1), ('heur/381328', 1), ('heur/381329', 1), ('heur/381330', 1), ('heur/381331', 1), ('heur/381332', 1), ('heur/381333', 1), ('heur/381334', 1), ('heur/383333', 1), ('heur/383334', 1), ('heur/383335', 1), ('heur/383336', 1), ('heur/383337', 1), ('heur/383338', 1), ('heur/383339', 1), ('heur/383340', 1), ('heur/383341', 1), ('heur/386342', 1), ('heur/386343', 1), ('heur/386344', 1), ('heur/386345', 1), ('heur/386346', 1), ('heur/386347', 1), ('heur/386348', 1), ('heur/386349', 1), ('heur/386350', 1), ('heur/386351', 1), ('heur/389350', 1), ('heur/389351', 1), ('heur/389352', 1), ('heur/389408', 1), ('heur/397415', 1), ('heur/397416', 1), ('heur/397417', 1), ('heur/397418', 1), ('heur/397419', 1), ('heur/397420', 1), ('heur/397421', 1), ('heur/397422', 1), ('heur/397423', 1), ('heur/397424', 1), ('heur/397425', 1), ('heur/397426', 1), ('heur/397427', 1), ('heur/397428', 1), ('heur/397429', 1), ('heur/397430', 1), ('heur/397431', 1), ('heur/397432', 1), ('heur/397433', 1), ('heur/397434', 1), ('heur/397435', 1), ('heur/397436', 1), ('heur/397437', 1), ('heur/397438', 1), ('heur/397439', 1), ('heur/397440', 1), ('heur/397441', 1), ('heur/397442', 1), ('heur/397443', 1), ('heur/397444', 1), ('heur/397445', 1), ('heur/397446', 1), ('heur/397447', 1), ('heur/397448', 1), ('heur/397449', 1), ('heur/397450', 1), ('heur/397451', 1), ('heur/397452', 1), ('heur/397453', 1), ('heur/397454', 1), ('heur/409455', 1), ('heur/409456', 1), ('heur/409457', 1), ('heur/409458', 1), ('heur/409459', 1), ('heur/409460', 1), ('heur/409461', 1), ('heur/409462', 1), ('heur/409463', 1), ('heur/409464', 1), ('heur/409465', 1), ('heur/409466', 1), ('heur/409467', 1), ('heur/409468', 1), ('heur/409469', 1), ('heur/409470', 1), ('heur/409471', 1), ('heur/409472', 1), ('heur/409473', 1), ('heur/409474', 1), ('heur/409475', 1), ('heur/409476', 1), ('heur/409477', 1), ('heur/409478', 1), ('heur/409479', 1), ('heur/409480', 1), ('heur/409481', 1), ('heur/409482', 1), ('heur/409483', 1), ('heur/409484', 1), ('heur/409485', 1), ('heur/409486', 1), ('heur/409487', 1), ('heur/409488', 1), ('heur/409489', 1), ('heur/409490', 1), ('heur/409491', 1), ('heur/421492', 1), ('heur/421493', 1), ('heur/421494', 1), ('heur/421495', 1), ('heur/421496', 1), ('heur/421497', 1), ('heur/421498', 1), ('heur/421499', 1), ('heur/421500', 1), ('heur/421501', 1), ('heur/421502', 1), ('heur/421503', 1), ('heur/421504', 1), ('heur/421505', 1), ('heur/421506', 1), ('heur/421507', 1), ('heur/421508', 1), ('heur/421509', 1), ('heur/421510', 1), ('heur/421511', 1), ('heur/421512', 1), ('heur/421513', 1), ('heur/421514', 1), ('heur/421515', 1), ('heur/421516', 1), ('heur/421517', 1), ('heur/421518', 1), ('heur/421519', 1), ('heur/421520', 1), ('heur/421521', 1), ('heur/421522', 1), ('heur/421523', 1), ('heur/421524', 1), ('heur/421525', 1), ('heur/421526', 1), ('heur/421527', 1), ('heur/421528', 1), ('heur/421529', 1), ('heur/421530', 1), ('heur/421531', 1), ('heur/433529', 1), ('heur/433530', 1), ('heur/433531', 1), ('heur/433532', 1), ('heur/433534', 1), ('heur/433535', 1), ('heur/433536', 1), ('heur/433537', 1), ('heur/433538', 1), ('heur/433539', 1), ('heur/433540', 1), ('heur/433541', 1), ('heur/445542', 1), ('heur/445543', 1), ('heur/445544', 1), ('heur/445545', 1), ('heur/445546', 1), ('heur/445547', 1), ('heur/445548', 1), ('heur/445549', 1), ('heur/445550', 1), ('heur/445551', 1), ('heur/445552', 1), ('heur/445553', 1), ('heur/445554', 1), ('heur/445555', 1), ('heur/457556', 1), ('heur/457557', 1), ('heur/457558', 1), ('heur/457559', 1), ('heur/457560', 1), ('heur/457561', 1), ('heur/457562', 1), ('heur/457563', 1), ('heur/457564', 1), ('heur/457565', 1), ('heur/457566', 1), ('heur/457567', 1), ('heur/457568', 1), ('heur/457569', 1), ('heur/469566', 1), ('heur/469567', 1), ('heur/469568', 1), ('heur/469569', 1), ('heur/469570', 1), ('heur/469571', 1), ('heur/469572', 1), ('heur/469573', 1), ('heur/469574', 1), ('heur/469575', 1), ('heur/469576', 1), ('heur/469577', 1), ('heur/469578', 1), ('heur/469579', 1), ('heur/471148', 1), ('heur/330959', 59)],


                                      }





#admissible_test = {"default": [fof_schedule[0]],
#                   0: [('heur/276653', 270)],
#                   1: [('heur/273498', 38), ('heur/277096', 183)],
#                   "cluster_centers": [[0, 0, 0], [1, 1, 1]],
#                   "model": "PATH_TO_MODEL"
#                   }
